'''
  $Id: helpers.py 192 2021-03-11 15:19:05Z cl_yubikey_2018a $
  
  Copyleft ITEG IT-Engineers GmbH 2022
'''

#import base64
#import configparser
#import glob
import os.path
import re
#import socket
import sys
import zipfile

#from xml.dom.minidom import parse
import xml.dom.minidom

from clazzes.jartools import helpers
from clazzes.jartools import jtlogging

log = jtlogging.getLogger()


class JTConfiguration:
    '''
      holds eventual parsed configuration and JAR file metadata
    '''

        # static holder for first instance
    staticJTConfiguration = None

    def __init__(self): 
        JTConfiguration.staticJTConfiguration = self
        self.jarMappings = {} # by pattern
        self.jarFiles = {} # by jar file name
        self.parseJarMapping()
        
    def parseJarMapping(self):
        if os.path.exists("/etc/jartools/jar-mappings.conf"):
            log.debug("Parsing /etc/jartools/jar-mappings.conf")
            jarMappingsDict = helpers.readConf("/etc/jartools/jar-mappings.conf")
            for jarPattern in sorted(jarMappingsDict.keys()):
                jarMapping = JARMapping(filenamePattern=jarPattern, mvnCoordinateTriple=jarMappingsDict[jarPattern])
                self.jarMappings[jarPattern] = jarMapping
            log.debug("Parsed %d JAR mappings" % len(self.jarMappings))
        else:
            log.warn("No JAR mapping to parse: No file /etc/jartools/jar-mappings.conf")


class JARMapping:

    def __init__(self, filenamePattern:str = "", mvnCoordinateTriple:str = ""):
        self.filenamePattern = filenamePattern
        self.compiledPattern = re.compile(filenamePattern)
        self.mvnCoordinateTriple = mvnCoordinateTriple
        self.mappedGroupId = None
        self.mappedArtifactId = None
        self.mappedVersion = None
        self.mappedName = None
        coordinateParts = self.mvnCoordinateTriple.split(":")
        nrCoordinateParts = len(coordinateParts)
        if nrCoordinateParts < 3:
            log.error("JARMapping(): JAR mapping '%s' has illegal coordinates specification (only %d ':'-separated parts): '%s'." % (
                self.filenamePattern, nrCoordinateParts, self.mvnCoordinateTriple))
            sys.exit(1)
        if nrCoordinateParts > 4:
            log.error("JARMapping(): JAR mapping '%s' has illegal coordinates specification (%d ':'-separated parts > 4): '%s'." % (
                self.filenamePattern, nrCoordinateParts, self.mvnCoordinateTriple))
            sys.exit(1)
        self.mappedGroupId = coordinateParts[0]
        self.mappedArtifactId = coordinateParts[1]
        self.mappedVersion = coordinateParts[2]
        if nrCoordinateParts > 3:
            self.mappedName = coordinateParts[3]
        else:
            self.mappedName = ""
        log.debug("C'ted JAR mapping from '%s' to '%s', i.e. groupId '%s', artifactId '%s', version '%s', name '%s'" % (
            self.filenamePattern, self.mvnCoordinateTriple,
            self.mappedGroupId, self.mappedArtifactId, self.mappedVersion, self.mappedName))


class JARFile:
    
    def __init__(self, path:str = "", jtConfiguration:JTConfiguration = None):
        self.path = path
        self.jtConfiguration = jtConfiguration
        self._clearCoordinates()

    def _clearCoordinates(self):
                # properties to get from pom xml or pom properties or jar mapping
        self.groupId = None
        self.artifactId = None
        self.version = None
        self.name = ""
        # status helper
        self.haveCoordinates = False

    def analyseJar(self):
        jarZipfile = zipfile.ZipFile(file=self.path, mode='r')
        pomPropertiesInZipPath = None
        pomXmlInZipPath = None
        nrPomPropertiesInZip = 0
        nrPomXmlInZip = 0
        for fileInZip in jarZipfile.namelist():
            #log.debug("JARFile.analyseJar[%s]: Analysing fileInZip %s" % (self.path, fileInZip))
            if not fileInZip.startswith("META-INF/maven/"):
                continue
            if fileInZip.endswith("/pom.properties"):
                nrPomPropertiesInZip += 1
                if pomPropertiesInZipPath == None:
                    pomPropertiesInZipPath = fileInZip
                #if not pomXmlInZipPath == None:
                #    break
                continue
            if fileInZip.endswith("/pom.xml"):
                nrPomXmlInZip += 1
                if pomXmlInZipPath == None:
                    pomXmlInZipPath = fileInZip
                #if not pomPropertiesInZipPath == None:
                #    break
                continue
        #
        if nrPomXmlInZip > 1:
            log.warning("JARFile.analyseJar[%s]: JAR has multiple META-INF/maven/**/pom.xml, not using any of them" % self.path)
            pomXmlInZipPath = None
        if nrPomPropertiesInZip > 1:
            log.warning("JARFile.analyseJar[%s]: JAR has multiple META-INF/maven/**/pom.properties, not using any of them" % self.path)
            pomPropertiesInZipPath = None
        # pom.xml?
        if pomXmlInZipPath == None:
            log.info("JARFile.analyseJar[%s]: JAR has no META-INF/maven/**/pom.xml or multiple thereof" % self.path)
        else:
            self.analysePomXml(jarZipfile=jarZipfile, fileInZip=pomXmlInZipPath)
            if self.haveCoordinates:
                return
            log.info("JARFile.analyseJar[%s]: JAR has a single META-INF/maven/**/pom.xml but that did not provide full coordinates" % self.path)
        # pom.properties?
        if pomPropertiesInZipPath == None:
            log.info("JARFile.analyseJar[%s]: JAR has no META-INF/maven/**/pom.properties or multiple thereof" % self.path)
        else:
            self.analysePomProperties(jarZipfile=jarZipfile, fileInZip=pomPropertiesInZipPath)
            if self.haveCoordinates:
                return
            log.info("JARFile.analyseJar[%s]: JAR has a single META-INF/maven/**/pom.properties but that did not provide full coordinates" % self.path)
        # jar mapping?
        self.attemptJarMapping()
        if self.haveCoordinates:
            return
        log.error("JARFile.analyseJar[%s]: attemptJarMapping() did not provide maven coordinates (and we had no usuable pom.properties or pom.xml)" % self.path)
        #sys.exit(1)
        
    def analysePomXml(self, jarZipfile:zipfile.ZipFile=None, fileInZip:str=None):
        log.debug("JARFile.analysePomXml[%s]: Reading %s" % (
            self.path, fileInZip))
        with jarZipfile.open(fileInZip, mode='r') as pomXmlFile:
            domTree = xml.dom.minidom.parse(pomXmlFile)
            xmlCollection = domTree.documentElement
        #for xmlObject in xmlCollection.getElementsByTagName("object"):
        self._clearCoordinates()
        parentNode = None
        parentGroupIdNode = None
        parentVersionNode = None
        groupIdNode = None
        artifactIdNode = None
        versionNode = None
        nameNode = None
        for pomNode in xmlCollection.childNodes:
            if pomNode.nodeType != pomNode.ELEMENT_NODE:
                continue
            if pomNode.nodeName == "parent":
                parentNode = pomNode
                continue
            if pomNode.nodeName == "groupId":
                groupIdNode = pomNode
                continue
            if pomNode.nodeName == "artifactId":
                artifactIdNode = pomNode
                continue
            if pomNode.nodeName == "version":
                versionNode = pomNode
                continue
            if pomNode.nodeName == "name":
                nameNode = pomNode
                continue
        if parentNode != None:
            for pomNode in parentNode.childNodes:
                if pomNode.nodeName == "groupId":
                    parentGroupIdNode = pomNode
                    continue
                if pomNode.nodeName == "version":
                    parentVersionNode = pomNode
                    continue
        # parse xml
        if groupIdNode != None:
            self.groupId = groupIdNode.firstChild.data
        elif parentGroupIdNode != None:
            self.groupId = parentGroupIdNode.firstChild.data
        if artifactIdNode != None:
            self.artifactId = artifactIdNode.firstChild.data
        if versionNode != None:
            self.version = versionNode.firstChild.data
        elif parentVersionNode != None:
            self.version = parentVersionNode.firstChild.data
        if nameNode != None:
            self.name = nameNode.firstChild.data
        # evaluate
        if self.groupId == None:
            log.error("JARFile.analysePomXml[%s]: %s did not provide a groupId" % (self.path, fileInZip))
            sys.exit(1)
        if self.artifactId == None:
            log.error("JARFile.analysePomXml[%s]: %s did not provide a artifactId" % (self.path, fileInZip))
            sys.exit(1)
        if self.version == None:
            log.error("JARFile.analysePomXml[%s]: %s did not provide a version" % (self.path, fileInZip))
            sys.exit(1)
        if self.name == None:
            log.info("JARFile.analysePomXml[%s]: %s did not provide a name" % (self.path, fileInZip))
        log.debug("JARFile.analysePomXml[%s]: %s did provide groupId '%s', artifactId '%s', version '%s', name '%s'" % (
            self.path, fileInZip, self.groupId, self.artifactId, self.version, self.name))
        self.haveCoordinates = True

    def analysePomProperties(self, jarZipfile:zipfile.ZipFile=None, fileInZip:str=None):
        log.debug("JARFile.analysePomProperties[%s]: Reading %s" % (
            self.path, fileInZip))
        with jarZipfile.open(fileInZip, mode='r') as pomPropertiesFile:
            for baLine in pomPropertiesFile.readlines():
                line = baLine.decode("utf-8") 
                if line[-1] == '\n':
                    line = line[:-1]
                if len(line) < 1:
                    continue
                if line[0] in "#;":
                    continue
                lineParts = line.split('=')
                if len(lineParts) < 1:
                    continue
                if len(lineParts) > 2:
                    continue
                if lineParts[0] == "groupId":
                    self.groupId = lineParts[1]
                    continue
                if lineParts[0] == "artifactId":
                    self.artifactId = lineParts[1]
                    continue
                if lineParts[0] == "version":
                    self.version = lineParts[1]
                    continue
                if lineParts[0] == "name":
                    self.name = lineParts[1]
                    continue
                log.warn("JARFile.analysePomProperties[%s]: %s has unknown property %s=%s" % (
                    self.path, fileInZip, lineParts[0], lineParts[1]))
        if self.groupId == None:
            log.error("JARFile.analysePomProperties[%s]: %s did not provide a groupId" % (self.path, fileInZip))
            sys.exit(1)
        if self.artifactId == None:
            log.error("JARFile.analysePomProperties[%s]: %s did not provide a artifactId" % (self.path, fileInZip))
            sys.exit(1)
        if self.version == None:
            log.error("JARFile.analysePomProperties[%s]: %s did not provide a version" % (self.path, fileInZip))
            sys.exit(1)
        #if self.name == None:
        #    self.name = "%s %s" % (self.groupId, self.artifactId)
        log.debug("JARFile.analysePomProperties[%s]: %s did provide groupId '%s', artifactId '%s', version '%s', name '%s'" % (
            self.path, fileInZip, self.groupId, self.artifactId, self.version, self.name))
        self.haveCoordinates = True
            
    def attemptJarMapping(self):
        baseName = os.path.basename(self.path)
        for jarMapping in self.jtConfiguration.jarMappings.values():
            match = jarMapping.compiledPattern.fullmatch(baseName)
            if match == None:
                continue
            #log.debug("JARFile.attemptJarMapping[%s]: Found jarMapping with pattern '%s'" % (
            #    self.path, jarMapping.filenamePattern))
            patchedMvnCoordinateTriple = re.sub(jarMapping.compiledPattern, jarMapping.mvnCoordinateTriple, baseName) 
            log.debug("JARFile.attemptJarMapping[%s]: Replacing '%s' in '%s' with '%s' gave us new coordinates '%s'" % (
                self.path, jarMapping.filenamePattern, baseName, jarMapping.mvnCoordinateTriple, patchedMvnCoordinateTriple))
            coordinateParts = patchedMvnCoordinateTriple.split(":")
            nrCoordinateParts = len(coordinateParts)
            if nrCoordinateParts < 3:
                log.error("JARFile.attemptJarMapping[%s]: JAR mapping '%s' has illegal coordinates specification (only %d ':'-separated parts): '%s' resp. '%s'." % (
                    self.path, jarMapping.filenamePattern, nrCoordinateParts, jarMapping.mvnCoordinateTriple, patchedMvnCoordinateTriple))
                sys.exit(1)
            if nrCoordinateParts > 4:
                log.error("JARFile.attemptJarMapping[%s]: JAR mapping '%s' has illegal coordinates specification (%d ':'-separated parts > 4): '%s' resp. '%s'." % (
                    self.path, jarMapping.filenamePattern, nrCoordinateParts, jarMapping.mvnCoordinateTriple, patchedMvnCoordinateTriple))
                sys.exit(1)
            self.groupId = coordinateParts[0]
            self.artifactId = coordinateParts[1]
            self.version = coordinateParts[2]
            if nrCoordinateParts > 3:
                self.name = coordinateParts[3]
            else:
                self.name = ""
            log.debug("JARFile.attemptJarMapping[%s]: JAR mapping '%s' gave us '%s', i.e. groupId '%s', artifactId '%s', version '%s', name '%s'" % (
                self.path, jarMapping.filenamePattern, patchedMvnCoordinateTriple,
                self.groupId, self.artifactId, self.version, self.name))
            self.haveCoordinates = True
     
    def describeCoordinates(self, descFormat:str=""):
        if descFormat == "text":
            return self._describeCoordinatesText()
        if descFormat == "txt":
            return self._describeCoordinatesTxt()
        if descFormat == "csv":
            return self._describeCoordinatesCsv()
        if descFormat == "pom.xml":
            return self._describeCoordinatesPomXml()
        log.error("JARFile.printCoordinates[%s]: Unsupported format '%s'" % descFormat)
        sys.exit(1)
        
    def _describeCoordinatesText(self):
        if not self.haveCoordinates:
            return "%s: -\n" % self.path
        desc = "%s:\n  groupId:    %s\n  artifactId: %s\n  version:    %s\n" % (
            self.path, self.groupId, self.artifactId, self.version)
        if self.name == None:
            return desc
        if len(self.name) == 0:
            return desc
        desc += "  name:       %s\n" % self.name
        return desc

    def _describeCoordinatesTxt(self):
        if not self.haveCoordinates:
            return "%s: -\n" % self.path
        desc = "%s: groupId %s, artifactId %s, version %s\n" % (
            self.path, self.groupId, self.artifactId, self.version)
        if self.name == None:
            return desc
        if len(self.name) == 0:
            return desc
        desc = "%s: groupId %s, artifactId %s, version %s, name %s\n" % (
            self.path, self.groupId, self.artifactId, self.version, self.name)
        return desc

    def _describeCoordinatesCsv(self):
        if not self.haveCoordinates:
            return "%s,,,,\n" % self.path
        if self.name == None:
            return "%s,%s,%s,%s,\n" % (
                self.path, self.groupId, self.artifactId, self.version)
        return "%s,%s,%s,%s,%s\n" % (
            self.path, self.groupId, self.artifactId, self.version, self.name)

    def _describeCoordinatesPomXml(self):
        desc =  "    <dependency>\n";
        # no real name support for dependencies, so do a comment
        # with syntax highlighting this looks better before the rest
        if (self.name != None) and (len(self.name) > 0):
            desc += "      <!-- <name>%s</name> -->\n" % self.name
        desc += "      <groupId>%s</groupId>\n" % self.groupId
        desc += "      <artifactId>%s</artifactId>\n" % self.artifactId
        desc += "      <version>%s</version>\n" % self.version
        desc += "    </dependency>\n"
        return desc
