'''
  $Id: helpers.py 192 2021-03-11 15:19:05Z cl_yubikey_2018a $

  Copyleft ITEG IT-Engineers GmbH 2022
'''

import fileinput
#import ipaddress
#import json
#import os
import re
#import socket
import subprocess
import sys
from functools import cmp_to_key

from clazzes.jartools import jtlogging

log = jtlogging.getLogger()


def formatFileSize(size:int = -1):
    if size == None:
        return "- (None)"
    if size < 0:
        return "- (< 0)"
    if size < 1024:
        return "%d B" % size
    if size < 1048576:
        return "%d KB" % (size/1024)
    if size < 1073741824:
        return "%d MB" % (size/1048576)
    if size < 1099511627776:
        return "%d GB" % (size/1073741824)
    return "%1.2f TB" % (size/1099511627776)


def parseFileSize(sizeString:str = ""):
    if sizeString == None:
        return -1
    if len(sizeString) == 0:
        return -2
    reInt = re.compile("^\d+$")
    #matchInt = reInt.search(sizeString)
    if reInt.fullmatch(sizeString):
        return int(sizeString)

    reIntUnit = re.compile("^(\d+)([mgtMGT])$")
    matchIntUnit = reIntUnit.fullmatch(sizeString)
    if matchIntUnit == None:
        return -3
    sizeInt = int(matchIntUnit.group(1))
    sizeBytes = -4
    if matchIntUnit.group(2) in "mM":
        sizeBytes = sizeInt * 1048576
    elif matchIntUnit.group(2) in "gG":
        sizeBytes = sizeInt * 1073741824
    elif matchIntUnit.group(2) in "tT":
        sizeBytes = sizeInt * 1099511627776
    #log.debug("parseFileSize(): Computed '%s' to %d" % (
    #    sizeString, sizeBytes))
    return sizeBytes


def readConf(confFileName:str = ""):
    if len(confFileName) == 0:
        raise Exception("ERROR: readConf() called with empty confFileName")
    conf = {}
    lineNr = 0
    for line in fileinput.input(files=(confFileName)):
        lineNr += 1
        if line[-1] == '\n':
            line = line[:-1]
        #log.debug("readConf(%s:%d): line='%s'" % (confFileName, lineNr, line))
        if len(line) == 0:
            #log.debug("readConf(%s:%d): Ignoring emtpy line" % (confFileName, lineNr))
            continue
        if line[0] in "#;":
            #log.debug("readConf(%s:%d): Comment line" % (confFileName, lineNr))
            continue
        lineParts = line.split('=')
        if len(lineParts) < 1:
            log.warning("readConf(%s:%d): Ignoring invalid line wihtout '=': '%s'" % (
                confFileName, lineNr, line))
            continue
        if len(lineParts[0]) == 0:
            log.warning("readConf(%s:%d): Ignoring invalid line without option name: '%s'" % (
                confFileName, lineNr, line))
            continue
        name = lineParts[0].strip()
        if len(name) == 0:
            log.warning("readConf(%s:%d): Ignoring invalid line with empty option name: '%s'" % (
                confFileName, lineNr, line))
            continue
        #log.debug("readConf(%s:%d): Trimmed name: '%s'" % (confFileName, lineNr, name))
        value = "=".join(lineParts[1:]).strip()
        # empty value is allowed
        #log.debug("readConf(%s:%d): Trimmed value: '%s'" % (confFileName, lineNr, value))
        if len(value) > 1 and value[0] == '"' and value[-1] == '"':
            value = value[1:-1]
        for key in conf.keys():
            value = value.replace("$%s" % key, conf[key]).replace("${%s}" % key, conf[key])
        #log.debug("readConf(%s:%d): $-replaced value: '%s'" % (confFileName, lineNr, value))
        conf[name] = value

    return conf


def backcheckOnProblem(rc:int=-1, cmdParts:[]=[], problem:str=""):
    if len(cmdParts) > 0:
        print("ATTENTION: The former command,\n  %s\nreturned the bad exit code %d." % (
            " ".join(cmdParts), rc), file=sys.stderr, flush=True)
    else:
        print("ATTENTION: The former command returned the bad exit code %d." %
              rc, file=sys.stderr, flush=True)

    if len(problem) > 0:
        print("PROBLEM description: %s" %
              problem, file=sys.stderr, flush=True)

    choice = input("Please try to correct the problem and re-execute the command manually in another terminal.\n"
                   "If your are You sure it is ok to continue, enter 'yes' ... ")
    if choice != "yes":
        log.error("You chose to ABORT halfway because of the problem mentioned above.")
        sys.exit(1)

    log.warning("Continuing on user's request")


def performDebugLogged(cmdParts:list=[], collectOutput=False):
    if len(cmdParts) == 0:
        raise Exception("ERROR: performDebugLogged() called with empty cmdParts")
    log.debug("Running\n  %s" % " ".join(cmdParts))
    cp = None
    if collectOutput:
        cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    else:
        cp = subprocess.run(cmdParts)
    rc = cp.returncode
    if rc != 0:
        log.debug("Got bad exitcode %d from %s" % (
            rc,  " ".join(cmdParts)))
    else:
        log.debug("Returncode was: %d" % rc)
    if collectOutput:
        return rc, cp.stdout, cp.stderr
    return rc

# https://maven.apache.org/pom.html#version-order-specification
def splitMavenVersion(a: str):
    # Implement maven version splitting rules.
    currentPartKind = None

    ret = []

    currentPart = ""
    currentSeparator = "."

    for char in a:
        if char in ("-", "."):
            ret.append((currentSeparator, currentPart))
            currentPart = ""
            currentPartKind = None
            currentSeparator = char
        elif char.isdigit() and currentPartKind == "alphabetic":
            ret.append((currentSeparator, currentPart))
            currentPart = char
            currentPartKind = "numeric"
            currentSeparator = "-"
        elif not char.isdigit() and currentPartKind == "numeric":
            ret.append((currentSeparator, currentPart))
            currentPart = char
            currentPartKind = "alphabetic"
            currentSeparator = "-"
        else:
            currentPart += char
            if currentPartKind is None:
                currentPartKind = "numeric" if char.isdigit() else "alphabetic"

    ret.append((currentSeparator, currentPart))

    return ret

def normalizeTokens(splitVersion):
    ret = []
    for (i, (separator, part)) in enumerate(splitVersion):
        nextNumeric = (i+1) < len(splitVersion) and splitVersion[i+1][1].isdigit()
        part = part.lower()
        if part == "":
            ret.append((separator, "0"))
        if part == "cr":
            ret.append((separator, "rc"))
        elif part == "a" and nextNumeric:
            ret.append((separator, "alpha"))
        elif part == "b" and nextNumeric:
            ret.append((separator, "beta"))
        elif part == "m" and nextNumeric:
            ret.append((separator, "milestone"))
        elif part in ("final", "ga", "release"):
            ret.append((separator, ""))
        else:
            ret.append((separator, part))

    return ret


def trimMavenVersion(splitVersion):
    reversedRet = []

    isTrimming = True

    for (separator, part) in reversed(splitVersion):
        if part in ("0", "", "final", "ga", "release") and isTrimming:
            pass
        else:
            reversedRet.append((separator, part))
            #isTrimming = separator == "-"
            isTrimming = not part.isdigit()

    reversedRet.reverse()
    return reversedRet

def padMavenVersions(v1, v2):
    if len(v1) == len(v2):
        return (v1,v2)

    (vl, vs) = (v1, v2) if len(v1) > len(v2) else (v2, v1)

    padLength = len(vl)

    for i in range(len(vs), padLength):
        (otherSeparator, otherPart) = vl[i]

        if not otherPart.isdigit():
        #if otherSeparator == "-":
            vs.append((otherSeparator, ""))
        else:
            vs.append((otherSeparator, "0"))

    return (v1,v2)

def cmp(a, b):
    return (a > b) - (a < b)

KNOWN_QUALIFIER = ["alpha", "beta", "milestone", "rc", "snapshot", "", "sp"]

def compareQualifier(q1, q2):
    if q1 in KNOWN_QUALIFIER and q2 in KNOWN_QUALIFIER:
        return cmp(KNOWN_QUALIFIER.index(q1), KNOWN_QUALIFIER.index(q2))
    elif q1 in KNOWN_QUALIFIER:
        return -1
    elif q2 in KNOWN_QUALIFIER:
        return 1
    else:
        return cmp(q1, q2)

def compareToken(t1, t2):
    (t1Separator, t1Part) = t1
    (t2Separator, t2Part) = t2

    if t1Separator == t2Separator:
        if t1Part.isdigit() and t2Part.isdigit():
            return cmp(int(t1Part), int(t2Part))
        elif t1Part.isdigit():
            return 1
        elif t2Part.isdigit():
            return -1
        else:
            return compareQualifier(t1Part, t2Part)
    else:
        if not t1Part.isdigit() and not t2Part.isdigit():
            return compareQualifier(t1Part, t2Part)
        elif t1Part.isdigit() and t2Part.isdigit():
            return cmp(t1Separator == ".", t2Separator == ".")
        else:
            return cmp(t1Part.isdigit(), t2Part.isdigit())


def parseMavenVersion(v: str):
    return trimMavenVersion(normalizeTokens(splitMavenVersion(v)))




def compareMavenVersions(a:str = None, b:str = None):
    if a == None:
        raise Exception("jartools compareMavenVersions('%s', '%s'): ERROR: a None" % (a, b))
    if b == None:
        raise Exception("jartools compareMavenVersions('%s', '%s'): ERROR: b None" % (a, b))
    if len(a) == 0:
        raise Exception("jartools compareMavenVersions('%s', '%s'): ERROR: a empty" % (a, b))
    if len(b) == 0:
        raise Exception("jartools compareMavenVersions('%s', '%s'): ERROR: b empty" % (a, b))

    v1 = parseMavenVersion(a)
    v2 = parseMavenVersion(b)

    (v1,v2) = padMavenVersions(v1,v2)

    keyFn = cmp_to_key(compareToken)

    return cmp(list(map(keyFn, v1)), list(map(keyFn, v2)))
