'''
  $Id: jt_bundle_lister.py 158 2021-02-18 20:35:39Z cl_yubikey_2018a $
  
  Copyleft ITEG IT-Engineers GmbH 2022
'''

import getopt
import glob
#import io
#import os
import os.path
#import subprocess
import sys

from clazzes.jartools import dto
from clazzes.jartools import jtlogging

log = jtlogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] jarFileOrDirectory...\n" % argv0
    usage += "\n"
    usage += "Lists bundles in specified JAR files and directories\n"
    usage += "based on JAR files' pom.properties and /etc/jartools/jar-mappings.conf.\n"
    usage += jtlogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    usage += "    -c csvFile  List in CSV format, separated with ','\n"
    usage += "    -i          Ignore (not abort over) missing files or directories\n"
    usage += "    -s          Use short text format, one line per jar\n"
    usage += "    -p pomSpec  List as pom.xml based on /etc/jartools/application-bundles_pom.xml\n"
    usage += "\n"
    usage += "For -p pomSpec = [applicationGroupId:applicationArtifactId:applicationVersion:applicationName:]pomToWrite.xml\n"
    usage += "If only pomToWrite.xml is specified the application coordinates and name are taken from the first JAR file.\n"
    usage += "If a JAR file is listed multiple time it's only analysed once.\n"
    usage += "\n"
    usage += "Before the content of the pom.xml template /etc/jartools/application-bundles_pom.xml is written\n"
    usage += "two replacements happen:\n"
    usage += "'<jt-application-coordinates/>' is replaced with the application's maven coordinates and name.\n"
    usage += "'<jt-dependencies/>' is replaced with the '<dependency>' tags.\n"
    #usage += "    -t    List tab separated.\n"
    return usage


def run_jt_bundle_lister(argv0,args):
    
    try:
        opts, jarPaths = getopt.getopt(args, "c:dhip:qstv", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    jtConfiguration = dto.JTConfiguration()

    listFormat = "text"
    csvFileName = None
    pomSpec = None
    ignoreMissingFiles = False

    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-c"]:
            listFormat = "csv"
            csvFileName = a
        elif o in ["-i"]:
            ignoreMissingFiles = True
        elif o in ["-p"]:
            listFormat = "pom.xml"
            pomSpec = a
        elif o in ["-s"]:
            listFormat = "txt"
        else:
            log.warning("Unhandled option: %s %s" %
                  (o, a))

    #if jarPaths = None:
    pomSpecParts = None
    if listFormat == "pom.xml":
        pomSpecParts = pomSpec.split(":")
        if len(pomSpecParts) == 1:
            log.notice("No app coordinates or name in pomSpec, going to read them from first JAR file.")
        elif len(pomSpecParts) != 5:
            log.error("Expecting 5 ':'-separated parts as pomSpec, but got %d parts." % len(pomSpecParts))
            sys.exit(1)

    if len(jarPaths) < 1:
        log.error("Expecting at least one jar file or directory.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)

    ''' check missing path parts '''
    firstJarPath = None
    for jarPath in jarPaths:
        if os.path.isfile(jarPath):
            log.debug("Found JAR file %s" % jarPath)
            jtConfiguration.jarFiles[jarPath] = dto.JARFile(path=jarPath, jtConfiguration=jtConfiguration)
            if firstJarPath == None:
                firstJarPath = jarPath
        elif os.path.isdir(jarPath):
            log.debug("Found JAR directory %s" % jarPath)
            for jarFileName in  glob.glob("%s/*.jar" % jarPath):
                log.debug("Found JAR file %s" % jarFileName)
                jtConfiguration.jarFiles[jarFileName] = dto.JARFile(path=jarFileName, jtConfiguration=jtConfiguration)
                if firstJarPath == None:
                    firstJarPath = jarFileName
        elif ignoreMissingFiles:
            log.debug("Ignoring missing jarPath '%s'" % jarPath)
        else:
            log.error("Aborting due to missing jarPath '%s'" % jarPath)
            sys.exit(1)

    log.info("Analysing %d jar files ..." % len(jtConfiguration.jarFiles))

    for jarFile in jtConfiguration.jarFiles.values():
        if jarFile.haveCoordinates:
            log.notice("No re-analysing jar file %s" % jarFile.path)
        jarFile.analyseJar()

    log.good("Analysed %d jar files." % len(jtConfiguration.jarFiles))
    
    listText = ""
    for jarFileName in sorted(jtConfiguration.jarFiles.keys()):
        jarFile = jtConfiguration.jarFiles[jarFileName]
        listText += jarFile.describeCoordinates(descFormat=listFormat)
    if listText[-1] == '\n':
        listText = listText[:-1]
        
    if listFormat in ("text", "txt"):
        print(listText, file=sys.stdout)
        sys.exit(0)
        
    if listFormat == "csv":
        with open(csvFileName, mode='w') as listFile:
            print("# JAR file,groupId,artifactId,version,name", file=listFile)
            print(listText, file=listFile)
            log.good("Wrote list to CSV file %s" % csvFileName)
        sys.exit(0)

    if not (listFormat == "pom.xml"):
        log.error("Unsupported format '%s'" % listFormat)
        sys.exit(1)

    # read pom template
    appCoordinates = ""
    if len(pomSpecParts) == 5:
        appCoordinates += "<groupId>%s</groupId>\n" % pomSpecParts[0]
        appCoordinates += "  <artifactId>%s</artifactId>\n" % pomSpecParts[1]
        appCoordinates += "  <version>%s</version>\n" % pomSpecParts[2]
        appCoordinates += "  <name>%s</name>" % pomSpecParts[3]
    elif firstJarPath == None:
        log.error("Got no application coordinates from pomSpec, and not first JAR file")
        sys.exit(1)
    else:
        ''' use from first JAR '''
        firstJarFile = jtConfiguration.jarFiles[firstJarPath]
        appCoordinates += "<groupId>%s</groupId>\n" % firstJarFile.groupId
        appCoordinates += "  <artifactId>%s</artifactId>\n" % firstJarFile.artifactId
        appCoordinates += "  <version>%s</version>\n" % firstJarFile.version
        if (firstJarFile.name == None) or (firstJarFile == ""):
            appCoordinates += "  <name>%s %s %s</name>" % (
                firstJarFile.groupId, firstJarFile.artifactId, firstJarFile.version)
        else:
            appCoordinates += "  <name>%s</name>" % firstJarFile.name
    pomToWriteFileName = pomSpecParts[-1]
        
    #pomTemplate = None
    with open("/etc/jartools/application-bundles_pom.xml", mode='r', encoding="utf-8") as pomTemplateFile:
        pomTemplate = pomTemplateFile.read()
    # replace header and dependencies
    filledPomText = pomTemplate.replace("<jt-application-coordinates/>", appCoordinates).replace("<jt-dependencies/>", listText)
    #filledPomText = pomTemplate.replace("<jt-dependencies/>", listText)
    with open(pomToWriteFileName, mode='w') as listFile:
        print(filledPomText, file=listFile)
        log.good("Wrote list to pom.xml file %s" % pomToWriteFileName)
    sys.exit(0)
