'''
  $Id: jt_regroup_bundles.py 158 2021-02-18 20:35:39Z cl_yubikey_2018a $
  
  Copyleft ITEG IT-Engineers GmbH 2022
'''

import getopt
import glob
#import io
#import os
import os.path
import shutil
#import subprocess
import sys

from clazzes.jartools import dto
from clazzes.jartools import helpers
from clazzes.jartools import jtlogging

log = jtlogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] jarFileOrDirectory... targetDirectory\n" % argv0
    usage += "\n"
    usage += "Regroups (merges) bundles from 2 or more directories into one new bundle\n"
    usage += "directory that will only have one version from each bundle.\n"
    usage += jtlogging.getCommonUsageOptions()
    usage += "  Specific options for version selection strategy are:\n"
    usage += "    -f     Use first occuring version of each bundle.\n"
    usage += "    -n     Use newest (highest) version of each bundle (default).\n"
    usage += "           Between euqal maven versions modification times are compared.\n"
    usage += "\n"
    usage += "  Specific options for file system behaviour are:\n"
    usage += "    -c     Copy JAR files instead of soft-linking them.\n"
    usage += "    -l     Use symlinks, don't copy JAR files (default).\n"
    usage += "\n"
    usage += "For bundles without a full triplet of maven coordinates a mapping is attempted\n"
    usage += "using /etc/jartools/jar-mappings.conf.\n"
    usage += "\n"
    usage += "If the targetDirectory does not exist it is created.\n"
    usage += "\n"
    usage += "If the targetDirectory already contains *.jar those are not removed or touched,\n"
    usage += "eventual smart checks might be added in the future.\n"
    usage += "\n"
    usage += "JAR files named '*-sources.jar' or '*-javadoc.jar' are ignored, allowing to use\n"
    usage += "target/ directories of maven projects as jarFileOrDirectory parameters.\n"
    #usage += "    -t    List tab separated.\n"
    return usage


def run_jt_regroup_bundles(argv0,args):
    
    try:
        opts, jarPathsTargetDirectory = getopt.getopt(args, "cdfhlnqtv", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    jtConfiguration = dto.JTConfiguration()

    versionStrategy = "newest"
    fsStrategy = "symlinking"

    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-c"]:
            fsStrategy = "copying"
        elif o in ["-f"]:
            versionStrategy = "first"
        elif o in ["-l"]:
            fsStrategy = "symlinking"
        elif o in ["-n"]:
            versionStrategy = "newest"
        else:
            log.warning("Unhandled option: %s %s" %
                  (o, a))

    if len(jarPathsTargetDirectory) < 2:
        log.error("Expecting at least one jar file or directory, and a target directory.\n")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)

    #helpers.testCompareMavenVersion()
    #sys.exit(99)
        
    jarPaths = jarPathsTargetDirectory[:-1]
    
    targetDirectory = jarPathsTargetDirectory[-1]
    if os.path.exists(targetDirectory):
        if os.path.isdir(targetDirectory):
            log.notice("Going to (re)fill existing target directory %s" % targetDirectory)
        else:
            log.error("Specified target directory %s exists, but not as directory" % targetDirectory)
            sys.exit(1)
    else:
        log.notice("Auto-creating target directory %s" % targetDirectory)
        os.makedirs(targetDirectory, mode=511)
        if not os.path.isdir(targetDirectory):
            log.error("Failed to auto-create target directory %s" % targetDirectory)
            sys.exit(1)

    # check missing path parts
    firstJarPath = None
    for jarPath in jarPaths:
        if os.path.isfile(jarPath):
            log.debug("Found JAR file %s" % jarPath)
            jtConfiguration.jarFiles[jarPath] = dto.JARFile(path=jarPath, jtConfiguration=jtConfiguration)
            if firstJarPath == None:
                firstJarPath = jarPath
        elif os.path.isdir(jarPath):
            log.debug("Found JAR directory %s" % jarPath)
            for jarFileName in  glob.glob("%s/*.jar" % jarPath):
                if jarFileName.endswith("-javadoc.jar"):
                    log.debug("Ignoring javadoc JAR %s" % jarFileName)
                    continue
                if jarFileName.endswith("-sources.jar"):
                    log.debug("Ignoring sources JAR %s" % jarFileName)
                    continue
                log.debug("Found JAR file %s" % jarFileName)
                jtConfiguration.jarFiles[jarFileName] = dto.JARFile(path=jarFileName, jtConfiguration=jtConfiguration)
                if firstJarPath == None:
                    firstJarPath = jarFileName
        else:
            log.warning("Ignoring missing jarPath '%s'" % jarPath)

    log.notice("Analysing %d jar files ..." % len(jtConfiguration.jarFiles))

    jarFilesByGroupArtifact = {}
    
    for jarFile in jtConfiguration.jarFiles.values():
        if jarFile.haveCoordinates:
            log.info("Not re-analysing or repeated jar file %s" % jarFile.path)
        jarFile.analyseJar()
        if not jarFile.haveCoordinates:
            jarFileDummyKey = os.path.basename(jarFile.path)
            if not (jarFileDummyKey in jarFilesByGroupArtifact.keys()):
                log.warning("In doubt using unidentified jar file %s, using basename %s as dummy key" % (jarFile.path, jarFileDummyKey))
                jarFilesByGroupArtifact[jarFileDummyKey] = jarFile
                continue
            log.warning("Ignoring 2nd unidentified jar file with already-known basename %s" % jarFileDummyKey)
            continue
        jarFileGroupArtifact = "%s:%s" % (jarFile.groupId, jarFile.artifactId)
        if not (jarFileGroupArtifact in jarFilesByGroupArtifact.keys()):
            log.info("File %s is the first with key %s, remembering JAR file." % (jarFile.path, jarFileGroupArtifact))
            jarFilesByGroupArtifact[jarFileGroupArtifact] = jarFile
            continue
        knownJarFile = jarFilesByGroupArtifact[jarFileGroupArtifact]
        if versionStrategy == "first":
            log.info("File %s is the 2nd with key %s after %s and version strategy is 'first', ignoring JAR file." % (
                jarFile.path, jarFileGroupArtifact, knownJarFile.path))
            continue
        if versionStrategy != "newest":
            log.error("Unknown versionStrategy '%s' while analysing file %s, skipping it" % (versionStrategy, jarFile.path))
            continue
        if jarFile.version == knownJarFile.version:
            log.debug("File %s is the 2nd with key %s and has the same version as %s, %s, comparing file creation times." % (
                jarFile.path, jarFileGroupArtifact, knownJarFile.path, jarFile.version))
            jarFileTS = os.path.getmtime(jarFile.path)
            knownJarFileTS = os.path.getmtime(knownJarFile.path)
            if jarFileTS > knownJarFileTS:
                log.info("File %s is the 2nd with key %s and has the same version as %s, %s, but is newer, using new JAR file only." % (
                    jarFile.path, jarFileGroupArtifact, knownJarFile.path, jarFile.version))
                jarFilesByGroupArtifact[jarFileGroupArtifact] = jarFile
                continue
            if jarFileTS < knownJarFileTS:
                log.info("File %s is the 2nd with key %s and has the same version as %s, %s, and is older, ignoring JAR file." % (
                    jarFile.path, jarFileGroupArtifact, knownJarFile.path, jarFile.version))
                continue
            if jarFileTS == knownJarFileTS:
                log.info("File %s is the 2nd with key %s and has the same version as %s, %s, and has identical modification mtime, ignoring JAR file." % (
                    jarFile.path, jarFileGroupArtifact, knownJarFile.path, jarFile.version))
                continue
            log.error("Reached unreachable point comparing modification timestamps of %s and %s" % (
                jarFile.path, knownJarFile.path))
        c = helpers.compareMavenVersions(jarFile.version, knownJarFile.version)
        if c > 0:
            log.info("File %s is the 2nd with key %s and has newer version than %s: %s > %s, using new JAR file." % (
                jarFile.path, jarFileGroupArtifact, knownJarFile.path, jarFile.version, knownJarFile.version))
            jarFilesByGroupArtifact[jarFileGroupArtifact] = jarFile
            continue
        if c < 0:
            log.info("File %s is the 2nd with key %s and has older version than %s: %s < %s, ignoring new JAR file." % (
                jarFile.path, jarFileGroupArtifact, knownJarFile.path, jarFile.version, knownJarFile.version))
            continue
        if c == 0:
            log.info("File %s is the 2nd with key %s and has equal-ish version with %s: %s == %s, ignoring new JAR file." % (
                jarFile.path, jarFileGroupArtifact, knownJarFile.path, jarFile.version, knownJarFile.version))
            continue
        log.error("Reached unreachable point comparing maven coordinates of %s and %s" % (
                jarFile.path, knownJarFile.path))

    log.notice("Analysed %d jar files." % len(jtConfiguration.jarFiles))
    
    for jarFileGroupArtifact in sorted(jarFilesByGroupArtifact.keys()):
        jarFile = jarFilesByGroupArtifact[jarFileGroupArtifact]
        targetFilename = "%s/%s" % (targetDirectory, os.path.basename(jarFile.path))
        if os.path.exists(targetFilename):
            log.notice("Not overwriting existing %s with copy of or link to %s" % (targetFilename, jarFile.path))
            continue
        if fsStrategy == "copying":
            log.notice("Copying %s to %s" % (jarFile.path, targetFilename))
            shutil.copy2(jarFile.path, targetFilename, follow_symlinks=True)
        elif fsStrategy == "symlinking":
            log.notice("Symlinking %s to %s" % (targetFilename, jarFile.path))
            os.symlink(jarFile.path, targetFilename)
        else:
            log.error("Unknown fsStrategy %s" % fsStrategy)
            sys.exit(1)

    log.good("Done merging (%s) %d JAR files into bundle directory %s with version strategy %s." % (
        fsStrategy, len(jarFilesByGroupArtifact), targetDirectory, versionStrategy))

    sys.exit(0)
